################################################################################
# Copyright (c) 2018-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import logging
import os
from typing import Iterable, Union, Optional, TypeVar

import redis

from .rdb_utility import encode_len
from .utils import ensure_binary, display_str, _PathType
from .telescope_state import TelescopeState
from .backend import Backend

# Basic RDB header. First 5 bytes are the standard REDIS magic
# Next 4 bytes store the RDB format version number (6 in this case)
# 0xFE flags the start of the DB selector (which is set to 0)
RDB_HEADER = b'REDIS0006\xfe\x00'

RDB_TERMINATOR = b'\xFF'

# Fast CRC-64 implementations on Python seem rare
# RDB is happy with zero checksum - we will protect the RDB dump in other ways
RDB_CHECKSUM = b'\x00\x00\x00\x00\x00\x00\x00\x00'

logger = logging.getLogger(__name__)

_T = TypeVar('_T', bound='RDBWriter')


def encode_item(key: bytes, dumped_value: bytes) -> bytes:
    """Encode key and corresponding DUMPed value to RDB format.

    The first byte indicates the encoding used for the value of this key. This
    is essentially just the Redis type. Subsequent bytes represent the key name,
    which consists of a length encoding followed by the actual byte
    representation of the string. Thereafter follows the value, encoded
    according to its appropriate schema. As a shortcut, the Redis DUMP command
    is used to generate the encoded value.

    Note: Redis provides a mechanism for optional key expiry, which we ignore here.
    """
    try:
        key_len = encode_len(len(key))
    except ValueError as exc:
        err = ValueError('Failed to encode key length: {}'.format(exc))
        raise err from None
    # The DUMPed value includes a leading type descriptor,
    # the encoded value itself (including length specifier),
    # a trailing version specifier (2 bytes) and finally an 8 byte checksum.
    # The version specified and checksum are discarded.
    key_type = dumped_value[:1]
    encoded_value = dumped_value[1:-10]
    return key_type + key_len + key + encoded_value


class RDBWriter:
    """RDB file resource that stores keys from one (or more) Redis DBs.

    Upon initialisation this opens the RDB file and writes the header.
    It is necessary to call :meth:`close` to write the trailer of the file
    and to close it properly (or use this object as a context manager).

    Parameters
    ----------
    filename : path-like
        Destination filename. Will be opened in 'wb'.

    Attributes
    ----------
    keys_written : int
        Number of keys written to the file.
    keys_failed : int
        Number of keys that failed to be written.
    """
    def __init__(self, filename: _PathType) -> None:
        self.filename = filename
        self.keys_written = 0
        self.keys_failed = 0
        self._fileobj = open(filename, 'wb')
        try:
            self._fileobj.write(RDB_HEADER)
        except Exception:
            # Make sure that file is at least closed if header failed
            self._fileobj.close()
            raise

    def __enter__(self: _T) -> _T:
        return self

    def __exit__(self, *exc_info) -> None:
        self.close()

    def close(self) -> None:
        """Close off RDB file and delete it if it contains no keys."""
        try:
            self._fileobj.write(RDB_TERMINATOR + RDB_CHECKSUM)
        finally:
            self._fileobj.close()
            if not self.keys_written:
                logger.error("No valid keys found - removing empty file")
                os.remove(self.filename)

    def save(self, client: Union[TelescopeState, Backend, redis.Redis],
             keys: Optional[Iterable[Union[str, bytes]]] = None) -> None:
        """Save a specified subset of keys from a Redis DB to the RDB dump file.

        This is a very limited RDB dump utility. It takes a Redis database
        either from a high-level :class:`~katsdptelstate.TelescopeState` object,
        its low-level backend or a compatible redis client. Missing or broken
        keys are skipped and logged without raising an exception.

        Parameters
        ----------
        client : :class:`~katsdptelstate.telescope_state.TelescopeState` or :class:`~katsdptelstate.telescope_state.Backend` or :class:`~redis.Redis`-like
            A telstate, backend, or Redis-compatible client instance supporting keys() and dump()
        keys : iterable of str or bytes, optional
            The keys to extract from Redis and include in the dump.
            Keys that don't exist will not raise an Exception, only a log message.
            None (default) includes all keys.
        """  # noqa: E501
        if isinstance(client, TelescopeState):
            client = client.backend
        if keys is None:
            logger.info("No keys specified - dumping entire database")
            keys = client.keys(b'*')
        for key in keys:
            key = ensure_binary(key)
            dumped_value = client.dump(key)
            try:
                if not dumped_value:
                    raise KeyError('Key not found in Redis')
                encoded_str = encode_item(key, dumped_value)
            except (ValueError, KeyError) as e:
                self.keys_failed += 1
                logger.error("Failed to save key %s: %s", display_str(key), e)
                continue
            self._fileobj.write(encoded_str)
            self.keys_written += 1


if __name__ == '__main__':
    import argparse
    import redis

    from .endpoint import endpoint_parser

    # Default Redis port for use in endpoint constructors
    DEFAULT_PORT = 6379

    logging.basicConfig()
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument('--outfile', default='simple_out.rdb', metavar='FILENAME',
                        help='Output RDB filename [default=%(default)s]')
    parser.add_argument('--keys', default=None, metavar='KEYS',
                        help='Comma-separated list of Redis keys to write to RDB. [default=all]')
    parser.add_argument('--redis', type=endpoint_parser(DEFAULT_PORT),
                        default=endpoint_parser(DEFAULT_PORT)('localhost'),
                        help='host[:port] of the Redis instance to connect to. '
                             '[default=%(default)s]')
    args = parser.parse_args()

    endpoint = args.redis
    logger.info("Connecting to Redis instance at %s", endpoint)
    client = redis.Redis(host=endpoint.host, port=endpoint.port)
    logger.info("Saving keys to RDB file %s", args.outfile)
    keys = args.keys
    if keys is not None:
        keys = keys.split(",")
    with RDBWriter(args.outfile) as rdb_writer:
        rdb_writer.save(client, keys)
    if rdb_writer.keys_failed > 0:
        logger.warning("Done - Warning %d keys failed to be written (%d succeeded)",
                       rdb_writer.keys_failed, rdb_writer.keys_written)
    else:
        logger.info("Done - %d keys written", rdb_writer.keys_written)



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import struct
import math
import functools
import enum
import os
import sys
from typing import Tuple, Union

import six


class KeyType(enum.Enum):
    IMMUTABLE = 1
    MUTABLE = 2
    INDEXED = 3


# Behave gracefully in case someone uses non-UTF-8 binary in a key on PY3
ensure_str = functools.partial(six.ensure_str, errors='surrogateescape')
ensure_binary = functools.partial(six.ensure_binary, errors='surrogateescape')
if sys.version_info >= (3, 6):
    _PathType = Union[bytes, str, os.PathLike]
else:
    _PathType = Union[bytes, str]


def display_str(s) -> str:
    """Return most human-readable and yet accurate version of *s*."""
    try:
        return '{!r}'.format(six.ensure_str(s))
    except UnicodeDecodeError:
        return '{!r}'.format(s)


def pack_query_timestamp(time: float, is_end: bool, include_end: bool = False) -> bytes:
    """Create a query value for a ZRANGEBYLEX query.

    When packing the time for the start of a range, set `is_end` and
    `include_end` to False. When packing the time for the end of a range,
    set `is_end` to True, and `include_end` indicates whether the endpoint
    is inclusive. The latter is implemented by incrementing the time by the
    smallest possible amount and then treating it as exclusive.
    """
    if time == math.inf:
        # The special positively infinite string represents the end of time
        return b'+'
    elif time < 0.0 or (time == 0.0 and not include_end):
        # The special negatively infinite string represents the dawn of time
        return b'-'
    else:
        packed_time = pack_timestamp(time)
        if include_end:
            # Increment to the next possible encoded value. Note that this
            # cannot overflow because the sign bit is initially clear.
            packed_time = struct.pack('>Q', struct.unpack('>Q', packed_time)[0] + 1)
        return (b'(' if is_end else b'[') + packed_time


def pack_timestamp(timestamp: float) -> bytes:
    """Encode a timestamp to a bytes that sorts correctly"""
    assert timestamp >= 0
    # abs forces -0 to +0, which encodes differently
    return struct.pack('>d', abs(timestamp))


def split_timestamp(packed: bytes) -> Tuple[bytes, float]:
    """Split out the value and timestamp from a packed item.

    The item contains 8 bytes with the timestamp in big-endian IEEE-754
    double precision, followed by the value.
    """
    assert len(packed) >= 8
    timestamp = struct.unpack('>d', packed[:8])[0]
    return (packed[8:], timestamp)



#!/usr/bin/env python

################################################################################
# Copyright (c) 2015-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import os.path

from setuptools import setup, find_packages


here = os.path.dirname(__file__)
readme = open(os.path.join(here, 'README.rst')).read()
news = open(os.path.join(here, 'NEWS.rst')).read()
long_description = readme + '\n\n' + news
tests_require = ['asynctest', 'async_timeout', 'fakeredis[lua,aioredis]>=1.4.0']

setup(name='katsdptelstate',
      description='Karoo Array Telescope - Telescope State Client',
      long_description=long_description,
      author='MeerKAT SDP team',
      author_email='sdpdev+katsdptelstate@ska.ac.za',
      packages=find_packages(),
      package_data={'': ['lua_scripts/*.lua', 'py.typed']},
      url='https://github.com/ska-sa/katsdptelstate',
      license='Modified BSD',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Astronomy'],
      platforms=['OS Independent'],
      keywords='meerkat ska',
      python_requires='>=3.5',
      setup_requires=['katversion'],
      use_katversion=True,
      install_requires=['redis>=3.3', 'six>=1.12', 'netifaces',
                        'msgpack', 'numpy'],
      extras_require={
          'rdb': ['rdbtools', 'python-lzf'],
          'aio': ['aioredis'],
          'test': tests_require
      },
      tests_require=tests_require,
      zip_safe=False     # For py.typed
      )



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from abc import ABC, abstractmethod
import time
from typing import List, Tuple, Dict, Generator, BinaryIO, Iterable, Optional, Union

from .utils import KeyType, _PathType


class KeyUpdateBase:
    """Indicates that the caller of the monitor should check again.

    This base class contains no information about what changed in the
    database. Sub-classes contain more specific information.
    """
    pass


class KeyUpdate(ABC, KeyUpdateBase):
    """Update notification for a specific key.

    This is a base class for type-specific notification classes and should not
    be instantiated directly.
    """

    def __init__(self, key: bytes, value: bytes) -> None:
        self.key = key
        self.value = value

    @property
    @abstractmethod
    def key_type(self) -> KeyType:
        pass       # noqa: cover


class MutableKeyUpdate(KeyUpdate):
    """Update notification for a mutable key."""

    def __init__(self, key: bytes, value: bytes, timestamp: float) -> None:
        super().__init__(key, value)
        self.timestamp = timestamp

    @property
    def key_type(self) -> KeyType:
        return KeyType.MUTABLE


class ImmutableKeyUpdate(KeyUpdate):
    """Update notification for an immutable key."""

    @property
    def key_type(self) -> KeyType:
        return KeyType.IMMUTABLE


class IndexedKeyUpdate(KeyUpdate):
    """Update notification for an indexed key."""

    def __init__(self, key: bytes, sub_key: bytes, value: bytes):
        super().__init__(key, value)
        self.sub_key = sub_key

    @property
    def key_type(self) -> KeyType:
        return KeyType.INDEXED


class Backend(ABC):
    """Low-level interface for telescope state backends.

    The backend interface does not deal with namespaces or encodings, which are
    handled by the frontend :class:`TelescopeState` class. A backend must be
    able to store the same types as :class:`~katsdptelstate.TelescopeState`,
    but keys and values will be :class:`bytes` rather than arbitrary Python
    objects.
    """

    @abstractmethod
    def load_from_file(self, file: Union[_PathType, BinaryIO]) -> int:
        """Implements :meth:`TelescopeState.load_from_file`."""

    @abstractmethod
    def __contains__(self, key: bytes) -> bool:
        """Return if `key` is in the backend."""

    @abstractmethod
    def keys(self, filter: bytes) -> List[bytes]:
        """Return all keys matching `filter`.

        The filter is a redis pattern. Backends might only support ``b'*'`` as
        a filter.
        """

    @abstractmethod
    def delete(self, key: bytes) -> None:
        """Delete a key (no-op if it does not exist)"""

    @abstractmethod
    def clear(self) -> None:
        """Remove all keys"""

    @abstractmethod
    def key_type(self, key: bytes) -> Optional[KeyType]:
        """Get type of `key`, or ``None`` if it does not exist."""

    @abstractmethod
    def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        """Set the value of an immutable key.

        If the key already exists (and is immutable), returns the existing
        value and does not update it. Otherwise, returns ``None``.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not immutable.
        """

    @abstractmethod
    def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        """Get the value and timestamp of a key.

        The return value depends on the key type:

        immutable
          The value.
        mutable
          The most recent value.
        indexed
          A dictionary of all values (with undefined iteration order).
        absent
          None

        The timestamp will be ``None`` for types other than mutable.
        """

    @abstractmethod
    def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        """Set a (value, timestamp) pair in a mutable key.

        The `timestamp` will be a non-negative float value.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not mutable
        """

    @abstractmethod
    def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        """Add value in an indexed immutable key.

        If the sub-key already exists, returns the existing value and does not
        update it. Otherwise, returns ``None``.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not indexed.
        """

    @abstractmethod
    def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        """Get the value of an indexed immutable key.

        Returns ``None`` if the key exists but the sub-key does not exist.

        Raises
        ------
        KeyError
            If the key does not exist.
        ImmutableKeyError
            If the key exists and is not indexed.
        """

    @abstractmethod
    def get_range(self, key: bytes, start_time: float, end_time: float,
                  include_previous: bool, include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        """Obtain a range of values from a mutable key.

        If the key does not exist, returns None.

        Parameters
        ----------
        key : bytes
            Key to search
        start_time : float
            Start of the range (inclusive).
        end_time : float
            End of the range. It is guaranteed to be non-negative.
        include_previous : bool
            If true, also return the last entry prior to `start_time`.
        include_end : bool
            If true, treat `end_time` as inclusive, otherwise exclusive.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not mutable
        """

    @abstractmethod
    def dump(self, key: bytes) -> Optional[bytes]:
        """Return a key in the same format as the Redis DUMP command, or None if not present."""

    def monitor_keys(self, keys: Iterable[bytes]) \
            -> Generator[Optional[KeyUpdateBase], Optional[float], None]:
        """Report changes to keys in `keys`.

        Returns a generator. The first yield from the generator is a no-op.
        After that, the caller sends a timeout and gets back an update event
        (of type :class:`KeyUpdateBase` or a subclass). If there is no event
        within the timeout, returns ``None``.

        It is acceptable (but undesirable) for this function to miss the
        occasional update e.g. due to a network connection outage. The caller
        takes care to use a low timeout and retry rather than blocking for
        long periods.

        The generator runs until it is closed.
        """
        # This is a valid but usually suboptimal implementation
        timeout = yield None
        while True:
            assert timeout is not None
            time.sleep(timeout)
            timeout = yield KeyUpdateBase()



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import time
import math
import logging
import contextlib
import warnings
from typing import List, Tuple, Dict, BinaryIO, Callable, Union, Optional, TypeVar, Any

import redis

from .endpoint import Endpoint, endpoint_parser
from .errors import (ImmutableKeyError, TimeoutError, CancelledError,
                     InvalidTimestampError)
from .encoding import ENCODING_DEFAULT, ENCODING_MSGPACK, encode_value, decode_value
from .utils import ensure_str, ensure_binary, display_str, KeyType, _PathType
from .backend import Backend, KeyUpdate, IndexedKeyUpdate
from .telescope_state_base import TelescopeStateBase, check_immutable_change


logger = logging.getLogger(__name__)
_Key = Union[bytes, str]
_T = TypeVar('_T', bound='TelescopeState')


class TelescopeState(TelescopeStateBase[Backend]):
    """Interface to attributes and sensors stored in a database.

    Refer to the README for a description of the types of keys supported.

    A Redis database used with this class must *only* be used with this class,
    as it assumes that all keys were encoded by this package. It should
    however be robust to malicious data, failing gracefully rather than
    executing arbitrary code or consuming unreasonable amounts of time or
    memory.

    Each instance of this class has an associated list of prefixes. Lookups
    try each key in turn until a match is found. Writes use the first prefix in
    the list. Conventionally, keys are arranged into a hierarchy, separated by
    underscores. A :meth:`view` convenience method helps with constructing
    prefix lists by automatically adding the trailing underscore to prefixes.

    Care should be used when attributes share a suffix. They may shadow
    each other for some views, causing the attribute to appear to have
    changed value. This class does not prevent it, because there is no way to
    know which namespaces may be shared in a view, and because doing it in a
    race-free way would be prohibitively expensive.

    Parameters
    ----------
    endpoint : str, :class:`~katsdptelstate.endpoint.Endpoint` or :class:`Backend`
        It can be

        - an endpoint: specifies the address of the Redis server

        - an URL (i.e., contains ``://``): passed to :meth:`redis.Redis.from_url`

        - an empty string: a :class:`~katsdptelstate.memory.MemoryBackend` is created

        - any other string: passed to :class:`~katsdptelstate.endpoint.Endpoint` to
          create an endpoint

        - a :class:`Backend`: used directly.
    db : int
        Database number within the Redis server.
    prefixes : tuple of str/bytes
        Prefixes that will be tried in turn for key lookup. While this can be
        specified directly for advanced cases, it is normally generated by
        :meth:`view`. Writes are made using the first prefix in the list.
    base : :class:`~katsdptelstate.telescope_state.TelescopeState`
        Existing telescope state instance, from which the backend will be
        taken. This allows new views to be created by specifying `prefixes`,
        without creating new backends.

    Raises
    ------
    ConnectionError
        If the initial connection to the (real) Redis server fails
    ValueError
        If a `base` is specified and either `endpoint` or `db` is non-default
    ValueError
        If `endpoint` is a :class:`Backend` and `db` is non-default
    """

    def __init__(self, endpoint: Union[str, Endpoint, Backend] = '',
                 db: int = 0, prefixes: Tuple[_Key, ...] = (b'',),
                 base: Optional['TelescopeState'] = None) -> None:
        if base is not None:
            if endpoint != '':
                raise ValueError('Cannot specify both base and endpoint')
            if db != 0:
                raise ValueError('Cannot specify both base and db')
            backend = None      # type: Optional[Backend]
        elif isinstance(endpoint, Backend):
            if db != 0:
                raise ValueError('Cannot specify both a backend and a db')
            backend = endpoint
        elif not endpoint:
            from .memory import MemoryBackend
            if db != 0:
                raise ValueError('Cannot specify a db when using the default backend')
            backend = MemoryBackend()
        else:
            from .redis import RedisBackend
            if isinstance(endpoint, str) and '://' in endpoint:
                r = redis.Redis.from_url(
                    endpoint,
                    db=db,
                    socket_timeout=5,
                    health_check_interval=30)
            else:
                if not isinstance(endpoint, Endpoint):
                    endpoint = endpoint_parser(default_port=None)(endpoint)
                assert isinstance(endpoint, Endpoint)   # Keeps mypy happy
                redis_kwargs = dict(
                    host=endpoint.host,
                    db=db,
                    socket_timeout=5,
                    health_check_interval=30)
                # If no port is provided, redis will pick its default port
                if endpoint.port is not None:
                    redis_kwargs['port'] = endpoint.port
                r = redis.Redis(**redis_kwargs)
            backend = RedisBackend(r)
        super().__init__(backend, prefixes, base)

    def load_from_file(self, file: Union[_PathType, BinaryIO]) -> int:
        """Load keys from a Redis-compatible RDB snapshot file.

        Redis keys are extracted sequentially from the RDB file and inserted
        directly into the backend without any checks and ignoring the view.
        It is therefore a bad idea to insert keys that already exist in telstate
        and this will lead to undefined behaviour. The standard approach is
        to call this method on an empty telstate.

        If there is an error reading or parsing the RDB file (indicating either
        a broken file or a non-RDB file), an `RdbParseError` is raised. Errors
        raised while opening the file (like `OSError`) and errors raised by the
        backend itself (like redis errors) can also occur.

        Parameters
        ----------
        file : str or file object
            Filename or file object representing RDB file

        Returns
        -------
        keys_loaded : int
            Number of keys loaded from RDB file into telstate

        Raises
        ------
        ImportError
            If the rdbtools package is not installed
        RdbParseError
            If the file could not be parsed (truncated / malformed / not RDB)
        """
        keys_loaded = self._backend.load_from_file(file)
        # mypy complains about {}-formatting of bytes
        logger.info("Loading {} keys from {}".format(keys_loaded, file))  # type: ignore
        return keys_loaded

    def __getattr__(self, key: str) -> Any:
        try:
            return self._get(key)
        except KeyError as error:
            raise AttributeError(str(error))

    def __getitem__(self, key: _Key) -> Any:
        return self._get(key)

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith('_'):
            super().__setattr__(key, value)
        elif any(key in cls.__dict__ for cls in self.__class__.__mro__):
            raise AttributeError("The specified key already exists as a class method "
                                 "and thus cannot be set via attribute access.")
        else:
            self.add(key, value, immutable=True)

    def __setitem__(self, key: _Key, value: Any) -> None:
        self.add(key, value, immutable=True)

    def __contains__(self, key: _Key) -> bool:
        """Check to see if the specified key exists in the database."""
        key = ensure_binary(key)
        for prefix in self._prefixes:
            if prefix + key in self._backend:
                return True
        return False

    def is_immutable(self, key: _Key) -> bool:
        """Check to see if the specified key is an immutable.

        Note that indexed keys are not considered immutable for this purpose.
        If the key does not exist, ``False`` is returned.

        .. deprecated:: 0.10
            :meth:`is_immutable` is deprecated and may be removed in a future release.
            Use :meth:`key_type` instead.
        """
        warnings.warn('is_immutable is deprecated; use key_type instead', FutureWarning)
        return self.key_type(key) == KeyType.IMMUTABLE

    def key_type(self, key: _Key) -> Optional[KeyType]:
        """Get the type of a key.

        If the key does not exist, returns ``None``.
        """
        key = ensure_binary(key)
        for prefix in self._prefixes:
            key_type = self._backend.key_type(prefix + key)
            if key_type is not None:
                return key_type
        return None

    def keys(self, filter: _Key = '*') -> List[str]:
        """Return a list of keys currently in the model.

        This function ignores the prefix list and returns all keys with
        fully-qualified names.

        Parameters
        ----------
        filter : str or bytes, optional
            Wildcard string passed to Redis to restrict keys

        Returns
        -------
        keys : list of str
            The key names, in sorted order
        """
        keys = self._backend.keys(ensure_binary(filter))
        return sorted(ensure_str(key) for key in keys)

    def _ipython_key_completions_(self) -> List[str]:
        """List of keys used in IPython (version >= 5) tab completion.

        This respects the prefix list and presents keys with prefixes removed.
        """
        keys = []     # type: List[str]
        keys_b = self._backend.keys(b'*')
        for prefix_b in self._prefixes:
            keys.extend(ensure_str(k[len(prefix_b):])
                        for k in keys_b if k.startswith(prefix_b))
        return keys

    def delete(self, key: _Key) -> None:
        """Remove a key, and all values, from the model.

        The key is deleted from every namespace in the prefix list.

        .. note::

            This function should be used rarely, ideally only in tests, as it
            violates the immutability of keys added with ``immutable=True``.
        """
        key = ensure_binary(key)
        for prefix in self._prefixes:
            self._backend.delete(prefix + key)

    def clear(self) -> None:
        """Remove all keys in all namespaces.

        .. note::

            This function should be used rarely, ideally only in tests, as it
            violates the immutability of keys added with ``immutable=True``.
        """
        return self._backend.clear()

    def add(self, key: _Key, value: Any, ts: Optional[float] = None,
            immutable: bool = False, encoding: bytes = ENCODING_DEFAULT) -> None:
        """Add a new key / value pair to the model.

        If `immutable` is true, then either the key must not previously have
        been set, or it must have been previously set immutable with exactly the
        same value (see :meth:`equal_encoded_values`). Thus, immutable keys only
        ever have one value for the lifetime of the telescope state. They also
        have no associated timestamp.

        Parameters
        ----------
        key : str or bytes
            Key name, which must not collide with a class attribute
        value : object
            Arbitrary value (must be encodable with `encoding`)
        ts : float, optional
            Timestamp associated with the update, ignored for immutables. If not
            specified, defaults to ``time.time()``.
        immutable : bool, optional
            See description above.
        encoding : bytes
            See :func:`encode_value`

        Raises
        ------
        ImmutableKeyError
            if an attempt is made to change the value of an immutable
        ImmutableKeyError
            if the key already exists and is not an immutable
        redis.ResponseError
            if there is some other error from the Redis server
        """
        key = ensure_binary(key)
        full_key = self._prefixes[0] + key
        key_str = display_str(full_key)
        str_val = encode_value(value, encoding)
        if immutable:
            try:
                old = self._backend.set_immutable(full_key, str_val)
            except ImmutableKeyError:
                raise ImmutableKeyError('Attempt to change key {} to immutable'
                                        .format(key_str))
            if old is not None:
                # The key already exists. Check if the value is the same.
                check_immutable_change(key_str, old, str_val, value)
        else:
            ts = float(ts) if ts is not None else time.time()
            if math.isnan(ts) or math.isinf(ts):
                raise InvalidTimestampError('Non-finite timestamps ({}) are not supported'
                                            .format(ts))
            if ts < 0.0:
                raise InvalidTimestampError('Negative timestamps ({}) are not supported'
                                            .format(ts))
            try:
                self._backend.add_mutable(full_key, str_val, ts)
            except ImmutableKeyError:
                raise ImmutableKeyError('Attempt to change key {} to mutable'
                                        .format(key_str))

    def set_indexed(self, key: _Key, sub_key: Any, value: Any,
                    encoding: bytes = ENCODING_DEFAULT) -> None:
        """Set a sub-key of an indexed key.

        Parameters
        ----------
        key : str or bytes
            Main key
        sub_key : object
            Sub-key within `key` to associate with the value. It must be both
            hashable and serialisable.
        encoding : bytes
            Encoding used for `value` (see :func:`encode_value`). Note that it
            does not affect the encoding of `sub_key`.

        Raises
        ------
        ImmutableKeyError
            if the sub-key already exists with a different value
        ImmutableKeyError
            if the key already exists and is not indexed
        redis.ResponseError
            if there is some other error from the Redis server
        """
        # Raises a TypeError if it's not hashable, to prevent trouble
        # retrieving it later.
        hash(sub_key)
        key = ensure_binary(key)
        full_key = self._prefixes[0] + key
        key_str = display_str(full_key)
        # Sub-keys will always be encoded with ENCODING_MSGPACK, so that
        # lookups don't need to worry whether they are using the matching
        # encoding.
        sub_key_enc = encode_value(sub_key, ENCODING_MSGPACK)
        str_val = encode_value(value, encoding)
        try:
            old = self.backend.set_indexed(full_key, sub_key_enc, str_val)
        except ImmutableKeyError:
            raise ImmutableKeyError('Attempt to change key {} to indexed immutable'
                                    .format(key_str))
        if old is not None:
            # The key already exists. Check if the value is the same.
            key_descr = '{}[{!r}]'.format(key_str, sub_key)
            check_immutable_change(key_descr, old, str_val, value)

    def get_indexed(self, key: _Key, sub_key: Any, default: Any = None,
                    return_encoded: bool = False) -> Any:
        """Retrieve an indexed value set with :meth:`set_indexed`.

        Parameters
        ----------
        key : str or bytes
            Main key
        sub_key : object
            Sub-key within `key`, which must be hashable and serialisable
        default : object
            Value to return if the sub-key is not found
        return_encoded : bool, optional
            Default 'False' - return values are first decoded from internal storage
            'True' - return values are retained in encoded form.
        """
        key = ensure_binary(key)
        sub_key_enc = encode_value(sub_key, ENCODING_MSGPACK)
        for prefix in self._prefixes:
            full_key = prefix + key
            try:
                raw_value = self.backend.get_indexed(full_key, sub_key_enc)
                if raw_value is None:
                    return default
                elif return_encoded:
                    return raw_value
                else:
                    return decode_value(raw_value)
            except KeyError:
                pass  # Key does not exist, try the next prefix
        return default

    def _check_condition(self, key: bytes,
                         condition: Optional[Callable[[Any, Optional[float]], bool]],
                         message: Optional[KeyUpdate] = None):
        """Check whether key exists and satisfies a condition (if any).

        Parameters
        ----------
        key : bytes
            Unqualified key name to check
        condition : callable, optional
            See :meth:`wait_key`'s docstring for the details
        message : :class:`.KeyUpdateBase`, optional
            A non-empty update returned by :meth:`.Backend.wait_key`.

            If specified, this is used to find the latest value and timestamp
            (if available) of the key instead of retrieving it from the backend.
        """
        if condition is None:
            return message is not None or key in self

        for prefix in self._prefixes:
            full_key = prefix + key
            if (message is not None and full_key == message.key
                    and message.key_type != KeyType.INDEXED):
                value = message.value     # type: Union[bytes, Dict[bytes, bytes]]
                timestamp = getattr(message, 'timestamp', None)    # type: Optional[float]
            else:
                value2, timestamp = self._backend.get(full_key)
                if value2 is None:
                    continue      # Key does not exist, so try the next one
                value = value2
            if isinstance(value, dict):
                # Handle indexed items
                value = {decode_value(k): decode_value(v) for k, v in value.items()}
            else:
                value = decode_value(value)
            return condition(value, timestamp)
        return False    # Key does not exist

    def wait_key(self, key: _Key,
                 condition: Optional[Callable[[Any, Optional[float]], bool]] = None,
                 timeout: Optional[float] = None,
                 cancel_future: Any = None) -> None:
        """Wait for a key to exist, possibly with some condition.

        Parameters
        ----------
        key : str or bytes
            Key name to monitor
        condition : callable, signature `bool = condition(value, ts)`, optional
            If not specified, wait until the key exists. Otherwise, the
            callable should have the signature `bool = condition(value, ts)`
            where `value` is the latest value of the key, `ts` is its
            associated timestamp (or None if immutable), and the return value
            indicates whether the condition is satisfied.
        timeout : float, optional
            If specified and the condition is not met within the time limit,
            an exception is thrown.
        cancel_future : future, optional
            If not ``None``, a future object (e.g.
            :class:`concurrent.futures.Future` or :class:`asyncio.Future`). If
            ``cancel_future.done()`` is true before the timeout, raises
            :exc:`CancelledError`. In the current implementation, it is only
            polled once a second, rather than waited for.

        Raises
        ------
        TimeoutError
            if a timeout was specified and was exceeded
        CancelledError
            if a cancellation future was specified and done
        """
        key = ensure_binary(key)
        # First check if condition is already satisfied, in which case we
        # don't need to create a pubsub connection.
        if self._check_condition(key, condition):
            return
        key_str = display_str(key)

        def check_cancelled():
            if cancel_future is not None and cancel_future.done():
                raise CancelledError('Wait for {} cancelled'.format(key_str))

        check_cancelled()
        monitor = self._backend.monitor_keys([prefix + key
                                              for prefix in self._prefixes])
        with contextlib.closing(monitor):
            monitor.send(None)   # Just to start the generator going
            start = time.time()
            while True:
                # redis-py automatically reconnects to the server if the connection
                # goes down, but we might miss messages in that case. So rather
                # than waiting an arbitrarily long time, we make sure to poll from
                # time to time. This also allows the cancellation future to be
                # polled.
                check_cancelled()
                get_timeout = 1.0
                if timeout is not None:
                    remain = (start + timeout) - time.time()
                    if remain <= 0:
                        raise TimeoutError('Timed out waiting for {} after {}s'
                                           .format(key_str, timeout))
                    get_timeout = min(get_timeout, remain)
                message = monitor.send(get_timeout)
                if message is None:
                    continue
                if not isinstance(message, KeyUpdate):
                    # The monitor thinks it's worth checking again, but doesn't
                    # have enough information to be useful.
                    message = None
                if self._check_condition(key, condition, message):
                    return

    def _check_indexed_condition(self, key: bytes, sub_key: bytes,
                                 condition: Optional[Callable[[Any], bool]],
                                 message: Optional[IndexedKeyUpdate] = None):
        """Check whether key exists and satisfies a condition (if any).

        Parameters
        ----------
        key : bytes
            Unqualified key name to check
        sub_key : bytes
            Encoded sub-key to check
        condition : callable, optional
            See :meth:`wait_indexed`'s docstring for the details
        message : :class:`.KeyUpdateBase`, optional
            A non-empty update returned by :meth:`.Backend.wait_key`. If
            specified, it must match the given `sub_key`.

            If specified, this is used to find the latest value instead of
            retrieving it from the backend.
        """
        assert message is None or message.sub_key == sub_key
        for prefix in self._prefixes:
            full_key = prefix + key
            if message is not None and full_key == message.key:
                value = message.value
            else:
                try:
                    # TODO: this could be more efficient if the backend
                    # provided a has_indexed that didn't retrieve the
                    # value.
                    value2 = self._backend.get_indexed(full_key, sub_key)
                except KeyError:
                    continue       # Key does not exist, try next prefix
                if value2 is None:
                    return False   # Key exists, but sub-key does not exist
                value = value2
            if not condition:
                return True
            else:
                return condition(decode_value(value))
        return False    # Key does not exist (for any prefix)

    def wait_indexed(self, key: _Key, sub_key: Any,
                     condition: Optional[Callable[[Any], bool]] = None,
                     timeout: Optional[float] = None, cancel_future: Any = None) -> None:
        """Wait for a sub-key of an indexed key to exist, possibly with some condition.

        Parameters
        ----------
        key : str or bytes
            Key name to monitor
        sub_key : object
            Sub-key to monitor within `key`.
        condition : callable, signature `bool = condition(value)`, optional
            If not specified, wait until the sub-key exists. Otherwise, the
            callable should have the signature `bool = condition(value)`
            where `value` is the value associated with the sub-key, and the
            return value indicates whether the condition is satisfied.
        timeout : float, optional
            If specified and the condition is not met within the time limit,
            an exception is thrown.
        cancel_future : future, optional
            If not ``None``, a future object (e.g.
            :class:`concurrent.futures.Future` or :class:`trollius.Future`). If
            ``cancel_future.done()`` is true before the timeout, raises
            :exc:`CancelledError`. In the current implementation, it is only
            polled once a second, rather than waited for.

        Raises
        ------
        TimeoutError
            if a timeout was specified and was exceeded
        CancelledError
            if a cancellation future was specified and done
        ImmutableKeyError
            if the key exists (or is created while waiting) but is not indexed
        """
        key = ensure_binary(key)
        sub_key_enc = encode_value(sub_key, ENCODING_MSGPACK)
        # First check if condition is already satisfied, in which case we
        # don't need to create a pubsub connection.
        if self._check_indexed_condition(key, sub_key_enc, condition):
            return
        key_str = '{}[{!r}]'.format(display_str(key), sub_key)

        def check_cancelled():
            if cancel_future is not None and cancel_future.done():
                raise CancelledError('Wait for {} cancelled'.format(key_str))

        check_cancelled()
        monitor = self._backend.monitor_keys([prefix + key
                                              for prefix in self._prefixes])
        with contextlib.closing(monitor):
            monitor.send(None)   # Just to start the generator going
            start = time.time()
            while True:
                check_cancelled()
                get_timeout = 1.0
                if timeout is not None:
                    remain = (start + timeout) - time.time()
                    if remain <= 0:
                        raise TimeoutError('Timed out waiting for {} after {}s'
                                           .format(key_str, timeout))
                    get_timeout = min(get_timeout, remain)
                message = monitor.send(get_timeout)
                if message is None:
                    continue
                elif not isinstance(message, KeyUpdate):
                    # The monitor thinks it's worth checking again, but doesn't
                    # have enough information to be useful.
                    message = None
                elif not isinstance(message, IndexedKeyUpdate):
                    raise ImmutableKeyError('wait_indexed called on non-indexed key {}'
                                            .format(key_str))
                if self._check_indexed_condition(key, sub_key_enc, condition, message):
                    return

    def _get(self, key: _Key, return_encoded: bool = False) -> Any:
        key = ensure_binary(key)
        for prefix in self._prefixes:
            full_key = prefix + key
            raw_value = self._backend.get(full_key)[0]
            if raw_value is not None:
                if return_encoded:
                    return raw_value
                elif isinstance(raw_value, dict):
                    # Indexed immutable
                    return {
                        decode_value(key): decode_value(value)
                        for (key, value) in raw_value.items()
                    }
                else:
                    return decode_value(raw_value)
        raise KeyError('{} not found'.format(display_str(key)))

    def get(self, key: _Key, default: Any = None, return_encoded: bool = False) -> Any:
        """Get a single value from the model.

        Parameters
        ----------
        key : str or bytes
            Key to retrieve
        default : object, optional
            Object to return if key not found
        return_encoded : bool, optional
            Default 'False' - return values are first decoded from internal storage
            'True' - return values are retained in encoded form.

        Returns
        -------
        value
            for non-immutable key return the most recent value
        """
        try:
            return self._get(key, return_encoded)
        except KeyError:
            return default

    def get_range(self, key: _Key,
                  st: Optional[float] = None, et: Optional[float] = None,
                  include_previous: Optional[bool] = None,
                  include_end: bool = False,
                  return_encoded: bool = False) -> List[Tuple[Any, float]]:
        """Get the range of values specified by the key and timespec from the model.

        Parameters
        ----------
        key : str or bytes
            Database key to extract
        st : float, optional
            Start time, default returns the most recent value prior to et
        et: float, optional
            End time, defaults to the end of time
        include_previous : bool, optional
            If True, the method also returns the last value
            prior to the start time (if any). This defaults to False if st is
            specified and True if st is unspecified.
        include_end : bool, optional
            If False (default), returns values in [st, et), otherwise [st, et].
        return_encoded : bool, optional
            Default 'False' - return values are first decoded from internal storage
            'True' - return values are retained in encoded form.

        Returns
        -------
        list
            list of (value, time) records in specified time range

        Raises
        ------
        KeyError
            if `key` does not exist (with any prefix)
        ImmutableKeyError
            if `key` refers to an existing key which is not mutable

        Notes
        -----
        By default, timestamps exactly equal to the start time are included,
        while those equal to the end time are excluded.

        Usage examples:

        get_range('key')
            returns most recent record

        get_range('key',st=0)
            returns list of all records in the telescope state database

        get_range('key',st=0,et=t1)
            returns list of all records before time t1

        get_range('key',st=t0,et=t1)
            returns list of all records in the range [t0,t1)

        get_range('key',st=t0)
            returns list of all records after time t0

        get_range('key',et=t1)
            returns the most recent record prior to time t1
        """
        # set up include_previous and st default values
        if include_previous is None:
            include_previous = True if st is None else False
        if et is None:
            et = math.inf
        else:
            et = float(et)
        if st is None:
            st = et
        else:
            st = float(st)
        if math.isnan(st) or math.isnan(et):
            raise InvalidTimestampError('cannot use NaN start or end time')

        key = ensure_binary(key)
        for prefix in self._prefixes:
            full_key = prefix + key
            try:
                ret_vals = self._backend.get_range(full_key, st, et,
                                                   include_previous, include_end)
            except ImmutableKeyError:
                raise ImmutableKeyError('{} is immutable, cannot use get_range'
                                        .format(display_str(full_key)))
            if ret_vals is not None:
                if not return_encoded:
                    ret_vals = [(decode_value(value), timestamp)
                                for value, timestamp in ret_vals]
                return ret_vals
        raise KeyError('{} not found'.format(display_str(key)))
